package xyz.tbvns.GeometryDash;

import com.badlogic.gdx.graphics.g3d.ModelInstance;
import com.badlogic.gdx.math.Vector3;
import lombok.extern.slf4j.Slf4j;
import org.joml.*;
import xyz.tbvns.Editor.ModelsManager;
import xyz.tbvns.GDConstant;
import xyz.tbvns.LevelInfo;
import xyz.tbvns.LevelStringBuilder;
import xyz.tbvns.Models.Face;
import xyz.tbvns.Models.GDInstance;
import xyz.tbvns.Models.GDModel;

import java.lang.Math;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import static xyz.tbvns.GDConstant.*;

@Slf4j
public class Render {
    public double currentPercent;
    public double currentDistance;
    public HashMap<GDInstance, Integer> animationToCurrentframe = new HashMap<>();

    public String fullRender(double fps) throws CloneNotSupportedException {
        LevelStringBuilder builder = new LevelStringBuilder(new StringBuilder(LevelInfo.getLevelString()), 10); //TODO: make the editor layer configurable
        Matrix4f camera = new Matrix4f();
        camera.perspective((float) Math.toRadians(2),  (float) 16 /9, 0.001F, 1000);
        while (currentPercent <= 1) {
            double speed = LevelInfo.getLevel().getSpeedAt((float) currentPercent);
            double distancePerFrame = (speed / fps);
            double realDistancePerFrame = (speed / fps) * 30;
            double distancePerFramePercent = realDistancePerFrame / LevelInfo.getLevel().getMaxPos() / 30;
            currentPercent += distancePerFramePercent;
            currentDistance += realDistancePerFrame;

            Vector3f cameraTranslation = new Vector3f((float) (currentPercent * LevelInfo.getLevel().getMaxPos()), 0, 5); //TODO: make that move with the player
            camera.translation(cameraTranslation); //TODO: make that follow a new camera path editor
            builder.add(1, 1, X_PROPERTY_ID, projectPoint(camera, new Vector4f(cameraTranslation.x, cameraTranslation.y, cameraTranslation.z - 1, 1)).x * 16 * 30, Y_PROPERTY_ID,projectPoint(camera, new Vector4f(cameraTranslation.x, cameraTranslation.y, cameraTranslation.z - 1, 1)).y * 9 * 30);

            System.out.println(currentPercent * 100);

            for (ModelInstance model : ModelsManager.getLoadedModels()) {
                GDInstance instance = (GDInstance) model.userData;
                if (instance.isAnimation()) {
                    //TODO: the code under here will cause issue with the first frame being drawn at the start of the level but not the others
                    int frame = instance.getFrameOnPercent(currentPercent);
                    if (animationToCurrentframe.containsKey(instance) && animationToCurrentframe.get(instance) != frame) {
                        animationToCurrentframe.put(instance, frame);
                        GDModel gdModel = instance.getAnimationObject().getFrames().get(frame);
                        for (Face face : gdModel.getFaces()) {
                            for (Vector4f point : face.points) {
                                Vector3 translation = instance.getInstance().transform.getTranslation(Vector3.Zero);
                                point = point.add(translation.x, translation.y, translation.z, 0);
                                Vector4f projected = projectPoint(camera, point);
                                builder.add(1, POINT_OBJECT_ID, X_PROPERTY_ID, projected.x * 16 * 30, Y_PROPERTY_ID, projected.y * 9 * 30);
                            }
                        }
                    } else {
                        animationToCurrentframe.put(instance, frame);
                    }
                } else {
                    GDModel gdModel = instance.getGdModel();
                    for (Face face : gdModel.getFaces()) {
                        for (Vector4f point : face.points) {
                            Vector4f vector = (Vector4f) point.clone();
                            Vector3 translation = instance.getInstance().transform.getTranslation(Vector3.Zero);
                            vector.add(translation.x * 30, translation.y * 30, translation.z * 30, 0);
                            Vector4f projected = projectPoint(camera, vector);
                            builder.add(1, POINT_OBJECT_ID, X_PROPERTY_ID, projected.x * 30, Y_PROPERTY_ID, projected.y * 30);
                        }
                    }
                }
            }
        }
        return builder.getLevelString().toString();
    }

    public Vector4f projectPoint(Matrix4f camera, Vector4f input) {
        input = camera.transform(input);

        input.x = input.x / input.z;
        input.y = input.y / input.z;
        input.z = input.z / input.z;

        return input;
    }

    public static float getZ(Face face) {
        //TODO: REDO ALL OF THIS WITH MY SAVIOUR "JOML" !
        //d = √[(x2 − x1)2 + (y2 − y1)2 + (z2 − z1)2]
        List<Double> dist = new ArrayList<>();

        face.points.forEach(p -> {
            dist.add((double) p.w);
        });

        double temp = 0;

        for (Double d : dist) {
            temp += d;
        }

        temp = temp / dist.size();

        return (float) temp;
    }
}
